/*
 *********************************************************************
 *                         N-Mice Simulation                         *
 *                                                                   *
 * Author: Arash Partow (1998)                                       *
 *                                                                   *
 * Copyright notice:                                                 *
 * Free use of the N-Mice Simulation code base is permitted under    *
 * the guidelines and in accordance with the most current version of *
 * the Common Public License.                                        *
 * http://www.opensource.org/licenses/cpl1.0.php                     *
 *                                                                   *
 *********************************************************************
*/


import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.imageio.ImageIO;
import java.io.*;
import java.math.*;
import java.util.*;

public class NMiceSim extends Applet
{

   final int  MAX_MICE   = 60;
   final int  _MICE      = 7;
   final long MAX_DELAY  = 100000;
   final int  MAX_X      = 500;
   final int  MAX_Y      = 500;
   final int  T_X        = MAX_X / 2;
   final int  T_Y        = MAX_Y / 2;

   int      MICE         = _MICE;
   Mouse[]  mouse        = new Mouse[MAX_MICE];
   Color[]  color        = new Color[MAX_MICE];
   Panel    drawPanel    = new Panel();
   Random   rand         = new Random();
   boolean  restartMode  = true;

   Image    dblBuffImg;
   Graphics dblBuffer;

   public void init()
   {
      setSize (MAX_X,MAX_Y);
      initialiseMice();

      dblBuffImg  = this.createImage(MAX_X,MAX_Y);
      dblBuffer   = dblBuffImg.getGraphics();

      AnimationThread at = new AnimationThread();
      at.delayedAnimation(this, 5);
      at.start();
   }

   public void paint(Graphics g)
   {
      if (restartMode)
      {
         dblBuffer.setColor(new Color(208,225,245));
         dblBuffer.fillRect(0,0,MAX_X,MAX_Y);
         drawShapeBoarder(dblBuffer);
         restartMode = false;
      }
      drawMice(dblBuffer);
      g.drawImage(dblBuffImg,0,0,null);
   }

   public void doMouseChase()
   {
      double dist;
      for (int i = 0; i < MICE; ++i)
      {
         dist = distance(mouse[i].proj.x(),mouse[i].proj.y(), mouse[(i + 1) % MICE].ox,mouse[(i + 1) % MICE].oy) * 0.02;
         mouse[i].mouseChase(mouse[(i + 1) % MICE].ox,mouse[(i + 1) % MICE].oy,dist);
      }
   }

   public void drawMice(Graphics g)
   {
      for (int i  = 0; i < MICE; ++i)
      {
         g.setColor(Color.black);
         mouse[i].drawMouse(g);
      }
   }

   public void initialiseMice()
   {
      double ang  = 360.0 / MICE;
      double rang = 0.0;
      double x    = 0.0;
      double y    = 0.0;
      int count   = 0;
      Projectile proj;

      for(int i = 0; i < MICE; ++i)
      {
         proj = new Projectile(0.0,0.0);
         proj.calcPath((360.0 - rang) % 360,(int)(MAX_X - 5 * (MAX_X * 0.105)));
         mouse[i] = new Mouse(proj.x(),proj.y(),T_X,T_Y);
         rang     = rang + ang;
         count++;
      }
   }

   public void drawShapeBoarder(Graphics g)
   {
      g.setColor(new Color(51,102,255));
      for(int i = 0; i < MICE; ++i)
      {
         g.drawLine((int)mouse[i].proj.x() + T_X,(int)(-1 * mouse[i].proj.y()) + T_Y,
                    (int)mouse[(i + 1) % MICE].proj.x() + T_X,(int)(-1 * mouse[(i + 1) % MICE].proj.y()) + T_Y);
      }
      g.setColor(new Color(208,225,245));
   }

   public double distance(double x1, double y1, double x2, double y2)
   {
      return (Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)));
   }

   public void animate()
   {
      if (Math.round(distance(mouse[0].proj.x(),mouse[0].proj.y(),mouse[1].proj.x(),mouse[1].proj.y())) > 2.0)
      {
         doMouseChase();
      }
      else
      {
         MICE = (Math.abs(rand.nextInt()) % 10) + 3;
         initialiseMice();
         restartMode = true;
      }
      this.paint(this.getGraphics());
   }

   class AnimationThread extends Thread
   {
      NMiceSim animationApplet;
      int      delay;

      public void delayedAnimation(NMiceSim a, int delay)
      {
         this.animationApplet = a;
         this.delay           = delay;
      }

      public void run()
      {
         while (true)
         {
            try
            {
               sleep(delay);
               animationApplet.animate();
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }
      }
   }
}


class Mouse
{
   Projectile proj;

   double ox    = 0.0;
   double oy    = 0.0;

   double destx = 0.0;
   double desty = 0.0;

   int T_X      = 0;
   int T_Y      = 0;

   public Mouse()
   {
      proj = new Projectile();
   }

   public Mouse(double x, double y)
   {
      proj = new Projectile(x,y);
      ox = x;
      oy = y;
   }

   public Mouse(double x, double y, int tx, int ty)
   {
      proj = new Projectile(x,y);
      ox   = x;
      oy   = y;
      T_X  = tx;
      T_Y  = ty;
   }

   public double x()
   {
      return proj.x();
   }

   public double y()
   {
      return proj.y();
   }

   public void mouseChase(double x, double y, double dist)
   {
      double bearing = getBearing(x,y);
      destx = x;
      desty = y;
      proj.calcPath(bearing,dist);
   }

   public double getBearing(double x, double y)
   {
      double nx   = x - proj.x();
      double ny   = y - proj.y();
      int    quad = quadrant(nx,ny);

      if (quad == 1 ) return Math.abs(Math.atan2( nx, ny) * 180.0 / Math.PI);
      if (quad == 2 ) return Math.abs(Math.atan2( ny,-nx) * 180.0 / Math.PI) + 270.0;
      if (quad == 3 ) return Math.abs(Math.atan2(-nx,-ny) * 180.0 / Math.PI) + 180.0;
      if (quad == 4 ) return Math.abs(Math.atan2(-ny, nx) * 180.0 / Math.PI) + 90.0;
      return 0;
   }

   public int quadrant(double x, double y)
   {
      if ((0.0 <= x) && (0.0 <= y)) return 1;
      if ((x <  0.0) && (0.0 <= y)) return 2;
      if ((x <  0.0) && (y <  0.0)) return 3;
      if ((0.0 <= x) && (y <  0.0)) return 4;
      return 0;
   }

   public void drawMouse(Graphics g)
   {
      Color color = g.getColor();
      g.setColor(new Color(51,102,255));
      g.fillOval((int)(ox - 3 + T_X),(int)(-oy - 3 + T_Y),6,6);
      g.setColor(color);
      g.fillOval((int)(proj.x() - 3 + T_X),(int)(-proj.y() - 3 + T_Y),6,6);
      ox = proj.x();
      oy = proj.y();
   }
}

class Projectile
{
   private PointXY  point;
   private double   currentBearing;

   public Projectile()
   {
      point = new PointXY();
      currentBearing = 0;
   }

   public Projectile(PointXY point)
   {
      this.point = point;
   }

   public Projectile(double x, double y)
   {
      point   = new PointXY();
      point.x = x;
      point.y = y;
   }

   public double x()
   {
      return point.x;
   }

   public double y()
   {
      return point.y;
   }

   public int calcPath(double bearing, double dist)
   {
      int xMult     = xMultiplier(bearing);
      int yMult     = yMultiplier(bearing);
      int quad      = quadrant(bearing);
      double rAng   = bearing % 90;
      double deltaX = 0.0;
      double deltaY = 0.0;

      if (quad % 2 == 0)
      {
         point.x = point.x + Math.cos(rAng * Math.PI / 180) * dist * xMult;
         point.y = point.y + Math.sin(rAng * Math.PI / 180) * dist * yMult;
      }
      else
      {
         point.x = point.x + Math.sin(rAng * Math.PI / 180.0) * dist * xMult;
         point.y = point.y + Math.cos(rAng * Math.PI / 180.0) * dist * yMult;
      }
      return 0;
   }

   public int quadrant(double ang)
   {
      if((ang >=   0) && (ang <  90)) return 1;
      if((ang >= 270) && (ang < 360)) return 2;
      if((ang >= 180) && (ang < 270)) return 3;
      if((ang >=  90) && (ang < 180)) return 4;
      return 0;
   }

   public int quadrant(double x, double y)
   {
      if ((0.0 <= x) && (0.0 <= y)) return 1;
      if ((x <  0.0) && (0.0 <= y)) return 2;
      if ((x <  0.0) && (y <  0.0)) return 3;
      if ((0.0 <= x) && (y <  0.0)) return 4;
      return 0;
   }

   public int xMultiplier(double ang)
   {
      switch (quadrant(ang))
      {
         case 1: return 1;
         case 2: return -1;
         case 3: return -1;
         case 4: return 1;
      }
      return 0;
   }

   public int yMultiplier(double ang)
   {
      switch (quadrant(ang))
      {
         case 1: return 1;
         case 2: return 1;
         case 3: return -1;
         case 4: return -1;
      }
      return 0;
   }

   public int deltaAngle(double ang)
   {
      switch (quadrant(ang))
      {
         case 1: return 0;
         case 2: return 270;
         case 3: return 180;
         case 4: return 90;
      }
      return -1;
   }

   public int deltaAngleQ(int quad)
   {
      switch (quad)
      {
         case 1: return 0;
         case 2: return 270;
         case 3: return 180;
         case 4: return 90;
      }
      return -1;
   }

   public double rotateX(double rang, double x, double y)
   {
      if (rang < 0.0) rang = 360.0 - rang;
      return x * Math.cos(rang * Math.PI / 180.0) - y * Math.sin(rang * Math.PI / 180.0);
   }

   public double rotateY(double rang, double x, double y)
   {
      if (rang < 0.0) rang = 360.0 - rang;
      return x * Math.sin(rang * Math.PI / 180.0) + y * Math.cos(rang * Math.PI / 180);
   }

}

class PointXY
{
   double x;
   double y;
   public PointXY()
   {
      x = 0.0;
      y = 0.0;
   }
}
